/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
 * Copyright (c)2001,2002 Adrian Drzewiecki
 */
#include "dzt.h"

typedef struct _NotebookInfo NotebookInfo;

struct _NotebookInfo
{
	gint ignore_click;
};

#define notebook_info(notebook)		GET_OBJECT_INFO(notebook, NotebookInfo)
#define notebook_ignore_click(notebook)	(notebook_info(notebook)->ignore_click)

static void notebook_show_or_hide_tabs (GtkWidget *notebook);

static gint on_notebook_button_press	(GtkWidget *notebook, GdkEventButton *ev);
static void on_notebook_switch_page	(GtkWidget *notebook);
static void on_notebook_realize		(GtkWidget *notebook);

static inline GtkWidget *
notebook_window(GtkWidget *notebook)
{
	return gtk_widget_get_ancestor(notebook, GTK_TYPE_WINDOW);
}

GtkWidget *
notebook_current_page(GtkWidget *notebook)
{
	GtkWidget *page;
	gint nr;

	nr   = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
	page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), nr);

	return page;
}

void
notebook_update_title(GtkWidget *notebook, gchar *title)
{
	window_update_title(notebook_window(notebook), title);
}

void
notebook_configure(GtkWidget *notebook)
{
	GtkWidget *page;
	gint       page_nr;

	page_nr = 0;
	for (;;) {
		page = gtk_notebook_get_nth_page(
				GTK_NOTEBOOK(notebook), page_nr ++);
		if (!page)
			break;

		page_configure(page);

		gtk_notebook_set_tab_label_packing(
				GTK_NOTEBOOK(notebook),
				page,
				GET_BOOL(notebook.expand_tabs),
				TRUE,
				GTK_PACK_START);
	}

	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), 
			GET_INT(notebook.tab_position));

	notebook_show_or_hide_tabs(notebook);

	gtk_notebook_set_tab_hborder(GTK_NOTEBOOK(notebook),
			GET_INT(notebook.title_padding));
}

void 
notebook_remove_page(GtkWidget *notebook, GtkWidget *page)
{
	gint nr;

	nr = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), page);
	gtk_notebook_remove_page(GTK_NOTEBOOK(notebook), nr);

	page = notebook_current_page(notebook);
	if (!page) {
		notebook_close(notebook);
		return;
	}

	notebook_show_or_hide_tabs(notebook);

	notebook_grab_focus(notebook);
}

void
notebook_set_ignore_click(GtkWidget *notebook)
{
	notebook_ignore_click(notebook) = TRUE;
}

void
notebook_grab_focus(GtkWidget *notebook)
{
	page_grab_focus(notebook_current_page(notebook));
}

void
notebook_go_prev_page(GtkWidget *notebook)
{
	gint pg = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
	if (pg == 0)
		while (gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), ++pg))
			;
	gtk_notebook_set_page(GTK_NOTEBOOK(notebook), pg - 1);
	notebook_grab_focus(notebook);
}

void
notebook_go_next_page(GtkWidget *notebook)
{
	gint pg = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook)) + 1;
	if (!gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), pg))
		pg = 0;
	gtk_notebook_set_page(GTK_NOTEBOOK(notebook), pg);
	notebook_grab_focus(notebook);
}

GtkWidget *
notebook_create()
{
	GtkWidget *nb;

	nb = gtk_notebook_new();

	gtk_notebook_set_scrollable(GTK_NOTEBOOK(nb), TRUE);

	gtk_signal_connect(GTK_OBJECT(nb), "button-press-event",
			GTK_SIGNAL_FUNC(on_notebook_button_press),(gpointer)nb);
	gtk_signal_connect(GTK_OBJECT(nb), "switch-page",
			GTK_SIGNAL_FUNC(on_notebook_switch_page), (gpointer)nb);
	gtk_signal_connect(GTK_OBJECT(nb), "realize",
			GTK_SIGNAL_FUNC(on_notebook_realize), (gpointer)nb);

	gtk_notebook_set_show_border(GTK_NOTEBOOK(nb), TRUE);
	gtk_notebook_set_tab_border(GTK_NOTEBOOK(nb), 0);
	gtk_widget_show(nb);

	notebook_configure(nb);
	return nb;
}

void
notebook_close(GtkWidget *notebook)
{
	window_close(notebook_window(notebook));
}

static void
notebook_show_or_hide_tabs(GtkWidget *notebook)
{
	gint show_tabs = TRUE;

	if (!GET_BOOL(notebook.single_terminal_has_tab))
			show_tabs = !page_is_lonely(notebook_current_page(notebook));

	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), show_tabs);
}

static gint
on_notebook_button_press(GtkWidget *notebook, GdkEventButton *ev)
{
	static GdkWindow *last_window = NULL;
	gboolean same_window;
	
	same_window = (last_window == ev->window);
	last_window = ev->window;

	if (notebook_ignore_click(notebook)) {
		notebook_ignore_click(notebook) = FALSE;
		notebook_grab_focus(notebook);
		return FALSE;
	}

	if (ev->button == 3) {
		menu_popup(notebook, ev->time);
		return TRUE;
	}

	if (ev->type == GDK_2BUTTON_PRESS && same_window) {
		GtkWidget *page = notebook_current_page(notebook);
		if (ev->window == page->window) {
			page_begin_rename(page, TRUE);
			return FALSE;
		}
	}

	return TRUE;
}

static void
on_notebook_switch_page(GtkWidget *notebook)
{
	notebook_grab_focus(notebook);
}

static void
on_notebook_realize(GtkWidget *notebook)
{
	GtkWidget	*app = gtk_widget_get_toplevel(notebook);
	GdkGeometry	hints;

	hints.base_width = (notebook->style->klass->xthickness * 2) + 2;
	hints.base_height = (notebook->style->klass->ythickness * 2);

	hints.width_inc = 1;
	hints.height_inc = 1;

	hints.min_width = hints.base_width + hints.width_inc;
	hints.min_height = hints.base_height + hints.height_inc;
		
	gtk_window_set_geometry_hints(GTK_WINDOW(app), notebook, &hints,
		GDK_HINT_RESIZE_INC|GDK_HINT_MIN_SIZE|GDK_HINT_BASE_SIZE);
}

