#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "prefs-gui-cb.h"
#include "prefs-gui.h"
#include "prefs-gui-sup.h"

#include "dzt.h"

typedef void (*ValFunc) (GtkWidget *, gchar *, ...);
typedef struct _ConfigItem ConfigItem;

struct _ConfigItem
{
	gchar name[PATH_MAX];
	ValFunc put_val;
	ValFunc get_val;
};

#define COMBO(name)	{ #name "_combo", (ValFunc)put_combo_val, (ValFunc)get_combo_val }
#define ENTRY(name)	{ #name "_entry", (ValFunc)put_entry_val, (ValFunc)get_entry_val }
#define COLOR(name)	{ #name "_picker", (ValFunc)put_color_val, (ValFunc)get_color_val }
#define CHECKBOX(name)	{ #name "_checkbox", (ValFunc)put_checkbox_val, (ValFunc)get_checkbox_val }
#define SPINNER(name)	{ #name "_spinner", (ValFunc)put_spinner_val, (ValFunc)get_spinner_val }
#define FONT(name)	{ #name "_picker", (ValFunc)put_font_val, (ValFunc)get_font_val }
#define FILENAME(name)	{ #name "_entry", (ValFunc)get_filename_val, (ValFunc)put_filename_val }
#define LAST_ITEM	{ "", NULL, NULL }

static GtkWidget *preferences_dialog = NULL;
static GtkWidget *profile_dialog = NULL;

static gchar *name_to_path(gchar *name);
static void   get_combo_val(GtkWidget *widget, gchar *name);
static void   put_combo_val(GtkWidget *widget, gchar *name);
static void   get_entry_val(GtkWidget *widget, gchar *name);
static void   put_entry_val(GtkWidget *widget, gchar *name);
static void   get_color_val(GtkWidget *widget, gchar *name);
static void   put_color_val(GtkWidget *widget, gchar *name);
static void   get_checkbox_val(GtkWidget *widget, gchar *name);
static void   put_checkbox_val(GtkWidget *widget, gchar *name);
static void   get_spinner_val(GtkWidget *widget, gchar *name);
static void   put_spinner_val(GtkWidget *widget, gchar *name);
static void   get_font_val(GtkWidget *widget, gchar *name);
static void   put_font_val(GtkWidget *widget, gchar *name);
static void   get_filename_val(GtkWidget *widget, gchar *name);
static void   put_filename_val(GtkWidget *widget, gchar *name);

static void   connect_changed(GtkWidget *widget, GtkWidget *dialog, gchar *signal);
static void   disconnect_changed(GtkWidget *widget);

static void   get_values(GtkWidget *ancestor, ConfigItem *items);
static void   put_values(GtkWidget *ancestor, ConfigItem *items);

static ConfigItem prefs_items[] = {
	COMBO(main_toolbar_style),
	CHECKBOX(main_debug_dnd),
	CHECKBOX(main_confirm_close_all),
	CHECKBOX(main_no_right_click_on_terminal),
	CHECKBOX(main_show_shortcuts),
	
	COMBO(notebook_tab_position),
	SPINNER(notebook_max_title_length),
	SPINNER(notebook_title_padding),
	CHECKBOX(notebook_single_terminal_has_tab),
	CHECKBOX(notebook_tabs_have_close_button),
	CHECKBOX(notebook_expand_tabs),

	ENTRY(shortcuts_menu),
	ENTRY(shortcuts_close_win),
	ENTRY(shortcuts_quit),
	ENTRY(shortcuts_next_tab),
	ENTRY(shortcuts_prev_tab),
	ENTRY(shortcuts_rename_tab),
	ENTRY(shortcuts_detach_tab),
	ENTRY(shortcuts_close_tab),
	ENTRY(shortcuts_prefs),

	LAST_ITEM
};

static ConfigItem profile_items[] = {
	ENTRY(program_name),
	ENTRY(program_command),
	COMBO(program_kill_signal),
	CHECKBOX(program_restart),
	CHECKBOX(program_change_notify),
	CHECKBOX(program_autostart),
	ENTRY(program_new_tab_key),
	ENTRY(program_new_win_key),

	COMBO(terminal_background_type),
	FILENAME(terminal_background_image),
	ENTRY(terminal_type),
	ENTRY(terminal_word_chars),
	FONT(terminal_font),
	CHECKBOX(terminal_enable_bell),
	CHECKBOX(terminal_del_is_del),
	CHECKBOX(terminal_blinking_cursor),
	CHECKBOX(terminal_update_title),
	CHECKBOX(terminal_swap_del),
	CHECKBOX(terminal_user_title),

	COMBO(scrolling_scrollbar_location),
	SPINNER(scrolling_scroll_buffer),
	CHECKBOX(scrolling_scroll_on_keystroke),
	CHECKBOX(scrolling_scroll_on_output),

	COLOR(colors_foreground),
	COLOR(colors_background),
	COLOR(colors_pal0),
	COLOR(colors_pal1),
	COLOR(colors_pal2),
	COLOR(colors_pal3),
	COLOR(colors_pal4),
	COLOR(colors_pal5),
	COLOR(colors_pal6),
	COLOR(colors_pal7),
	COLOR(colors_pal8),
	COLOR(colors_pal9),
	COLOR(colors_pal10),
	COLOR(colors_pal11),
	COLOR(colors_pal12),
	COLOR(colors_pal13),
	COLOR(colors_pal14),
	COLOR(colors_pal15),
	COLOR(colors_change_notify),
	CHECKBOX(colors_invert),

	LAST_ITEM
};

void
edit_prefs()
{
	if (!preferences_dialog)
		preferences_dialog = create_preferences_dialog();

	gtk_widget_show_all(preferences_dialog);
}

static gint
get_selected_nr()
{
	if (preferences_dialog) {
		GtkWidget *clist = lookup_widget(preferences_dialog, "profile_list");
		return (gint)gtk_object_get_data(GTK_OBJECT(clist), "selected");
	}
	return 0;
}

static void
set_selected_nr(gint nr)
{
	if (nr >= 0 && preferences_dialog) {
		GtkWidget *clist = lookup_widget(preferences_dialog, "profile_list");
		gtk_clist_unselect_all(GTK_CLIST(clist));
		gtk_clist_select_row(GTK_CLIST(clist), nr, 0);
	}
}

static void
update_profile_list()
{
	if (preferences_dialog) {
		GtkWidget *clist = lookup_widget(preferences_dialog, "profile_list");
		gint reselect = get_selected_nr();
		gint nr = 0;
		
		while(push_profile_nr(nr)) {
			gchar *name = GET_STR(program.name);
			gchar *command = GET_STR(program.command);
			gboolean restart = GET_BOOL(program.restart);
			gchar *text[3] = { name, command, restart ? "yes" : "no" };

			gtk_clist_insert(GTK_CLIST(clist), nr, text);
			gtk_clist_remove(GTK_CLIST(clist), nr + 1);

			g_free(name);
			g_free(command);
			pop_profile();

			nr ++;
		}

		while (gtk_clist_get_text(GTK_CLIST(clist), nr, 0, NULL))
			gtk_clist_remove(GTK_CLIST(clist), nr);

		set_selected_nr(reselect);
	}
}

static void
emit_signal(GtkWidget *widget, gchar *signal)
{
	gtk_signal_emit_by_name(GTK_OBJECT(widget), signal, NULL);
}

static void
update_profile_dialog()
{
	if (profile_dialog) {
		if (get_selected_nr() >= 0) {
			push_profile_nr(get_selected_nr());
			get_values(profile_dialog, profile_items);
			pop_profile();
			gtk_widget_set_sensitive(GTK_WIDGET(profile_dialog), TRUE);
		} else {
			gtk_widget_set_sensitive(GTK_WIDGET(profile_dialog), FALSE);
		}
	}
}

void
update_preferences_dialog()
{
	if (preferences_dialog) {
		get_values(preferences_dialog, prefs_items);
		update_profile_list();
		update_profile_dialog();
	}
}

void
on_profile_list_select_row             (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
	gtk_object_set_data(GTK_OBJECT(clist), "selected", (gpointer)row);

	if (event && event->type == GDK_2BUTTON_PRESS) {
		if (!profile_dialog)
			profile_dialog = create_profile_dialog();

		gtk_widget_show_all(profile_dialog);
	}

	update_profile_dialog();
}


void
on_profile_list_unselect_row           (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
	gtk_object_set_data(GTK_OBJECT(clist), "selected", (gpointer)-1);
	update_profile_dialog();
}


gboolean
on_profile_list_key_press_event        (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data)
{
	if (get_selected_nr() >= 1) {
		GtkWidget *window = gtk_widget_get_toplevel(widget);
		GtkWidget *delete_profile = lookup_widget(window, "delete_profile");
		GtkWidget *edit_profile = lookup_widget(window, "edit_profile");

		switch(event->keyval) {
			case GDK_Delete:
			case GDK_BackSpace:
				emit_signal(delete_profile, "clicked");
				break;
			case GDK_KP_Enter:
			case GDK_ISO_Enter:
			case GDK_3270_Enter:
			case GDK_Return:
				emit_signal(edit_profile, "clicked");
				break;
		}

		return FALSE;
	}

	return TRUE;
}


void
on_profile_dialog_show                 (GtkWidget       *widget,
                                        gpointer         user_data)
{
	update_profile_dialog();
}


void
on_profile_dialog_apply                (GnomePropertyBox *gnomepropertybox,
                                        gint             arg1,
                                        gpointer         user_data)
{
	if (get_selected_nr() >= 0) {
		push_profile_nr(get_selected_nr());
		put_values(GTK_WIDGET(gnomepropertybox), profile_items);
		pop_profile();
		config_sync();
	}
}


void
on_preferences_dialog_apply            (GnomePropertyBox *gnomepropertybox,
                                        gint             arg1,
                                        gpointer         user_data)
{
	put_values(GTK_WIDGET(gnomepropertybox), prefs_items);
	config_sync();
}


void
on_profile_dialog_destroy              (GtkObject       *object,
                                        gpointer         user_data)
{
	profile_dialog = NULL;
}

void
on_preferences_dialog_show             (GtkWidget       *widget,
                                        gpointer         user_data)
{
	update_preferences_dialog();
}

static gchar *
name_to_path(gchar *name)
{
	name = g_strdup(name);
	*(strchr(name, '_')) = '.';
	*(strrchr(name, '_')) = '\0';

	return name;
}

static void
get_combo_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gint num = gnome_config_get_int(path);
	g_free(path);

	disconnect_changed(widget);
	gtk_list_select_item(GTK_LIST(GTK_COMBO(widget)->list), num);
	connect_changed(GTK_COMBO(widget)->entry, ancestor, "changed");
}

static void
put_combo_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	GtkList *list = GTK_LIST(GTK_COMBO(widget)->list);
	if (list->selection) {
		gchar *path = name_to_path(name);
		gint value = g_list_index(list->children, list->selection->data);
		gnome_config_set_int(path, value);
		g_free(path);
	}
}

static void
get_entry_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gchar *value = gnome_config_get_string(path);
	g_free(path);

	disconnect_changed(widget);
	if (value) {
		gtk_entry_set_text(GTK_ENTRY(widget), value);
		g_free(value);
	}
	connect_changed(widget, ancestor, "changed");
}

static void
put_entry_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gchar *value = gtk_editable_get_chars(GTK_EDITABLE(widget), 0, -1);
	gnome_config_set_string(path, value);
	g_free(path);
	g_free(value);
}

static void
get_spinner_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gint num = gnome_config_get_int(path);
	g_free(path);

	disconnect_changed(widget);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), num);
	connect_changed(widget, ancestor, "changed");
}

static void
put_spinner_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gint value = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
	gnome_config_set_int(path, value);
	g_free(path);
}

static void
get_checkbox_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gint num = gnome_config_get_bool(path);
	g_free(path);

	disconnect_changed(widget);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), num);
	connect_changed(widget, ancestor, "toggled");
}

static void
put_checkbox_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gboolean value = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
	gnome_config_set_bool(path, value);
	g_free(path);
}

static void
get_color_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gchar *value = gnome_config_get_string(path);
	gchar *ptr = value;
	gushort color[4] = {
		(gushort)strtol(ptr, &ptr, 0),
		(gushort)strtol(ptr, &ptr, 0),
		(gushort)strtol(ptr, &ptr, 0),
		(gushort)0xffff
	};
	g_free(path);

	disconnect_changed(widget);
	gnome_color_picker_set_i16(GNOME_COLOR_PICKER(widget),
			color[0], color[1], color[2], color[3]);

	connect_changed(widget, ancestor, "color-set");

	g_free(value);
}

static void
put_color_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gchar *value;
	gushort color[4];

	gnome_color_picker_get_i16(GNOME_COLOR_PICKER(widget),
			&color[0], &color[1], &color[2], &color[3]);

	value = g_strdup_printf("0x%04x 0x%04x 0x%04x", color[0], color[1], color[2]);

	gnome_config_set_string(path, value);

	g_free(path);
	g_free(value);
}

static void
get_font_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gchar *value = gnome_config_get_string(path);

	disconnect_changed(widget);
	gnome_font_picker_set_font_name(GNOME_FONT_PICKER(widget), value);
	connect_changed(widget, ancestor, "font_set");

	g_free(value);
	g_free(path);
}

static void
put_font_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gchar *value = gnome_font_picker_get_font_name(GNOME_FONT_PICKER(widget));

	gnome_config_set_string(path, value);

	g_free(path);
}

static void
get_filename_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gchar *value = gnome_config_get_string(path);

	disconnect_changed(gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(widget)));
	gnome_file_entry_set_default_path(GNOME_FILE_ENTRY(widget), value);
	connect_changed(gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(widget)), ancestor, "changed");

	g_free(value);
	g_free(path);
}

static void
put_filename_val(GtkWidget *ancestor, gchar *name)
{
	GtkWidget *widget = lookup_widget(ancestor, name);
	gchar *path = name_to_path(name);
	gchar *value = gnome_file_entry_get_full_path(GNOME_FILE_ENTRY(widget), TRUE);

	gnome_config_set_string(path, value);

	g_free(value);
	g_free(path);

}

static void
get_values(GtkWidget *ancestor, ConfigItem *items)
{
	while(items->get_val) {
		(*items->get_val)(ancestor, items->name);
		items++;
	}
}

static void
put_values(GtkWidget *ancestor, ConfigItem *items)
{
	while(items->put_val) {
		(*items->put_val)(ancestor, items->name);
		items++;
	}
}

static void
connect_changed(GtkWidget *widget, GtkWidget *dialog, gchar *signal)
{
	guint id = (guint)gtk_object_get_data(GTK_OBJECT(widget), "$changed_signal");
	if (!id) {
		id = gtk_signal_connect_object(GTK_OBJECT(widget),
					       signal,
					       GTK_SIGNAL_FUNC(gnome_property_box_changed),
					       (gpointer)dialog);
		gtk_object_set_data(GTK_OBJECT(widget), "$changed_signal", (gpointer)id);
	}
}

static void
disconnect_changed(GtkWidget *widget)
{
	guint id = (guint)gtk_object_get_data(GTK_OBJECT(widget), "$changed_signal");
	if (id) {
		gtk_signal_disconnect(GTK_OBJECT(widget), id);
		gtk_object_set_data(GTK_OBJECT(widget), "$changed_signal", (gpointer)0);
	}
}

void
on_change_shortcut_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
	GtkWidget *entry = GTK_WIDGET(button);
	gtk_widget_set_sensitive(entry, TRUE);
	gtk_widget_grab_focus(entry);
	gtk_editable_select_region(GTK_EDITABLE(entry), 0, -1);
}


gboolean
on_shortcut_key_pressed                (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data)
{
        gchar *accel;

        switch(event->keyval) {
                case GDK_Delete:
                case GDK_BackSpace:
                        gtk_entry_set_text(GTK_ENTRY(widget), "");
			gtk_widget_set_sensitive(widget, FALSE);
                        break;
                default:
                        event->state &= ~GDK_LOCK_MASK; // Ignore Caps Lock
                        event->state &= ~GDK_MOD2_MASK; // Ignore Num Lock
                        accel = gtk_accelerator_name(event->keyval, event->state);
                        if (gtk_accelerator_valid(event->keyval, event->state)) {
                                gtk_entry_set_text(GTK_ENTRY(widget), accel);
				gtk_widget_set_sensitive(widget, FALSE);
                        } 
        }
        return TRUE;
}


gboolean
on_shortcut_unfocus                    (GtkWidget       *widget,
                                        GdkEventFocus   *event,
                                        gpointer         user_data)
{
	gtk_widget_set_sensitive(widget, FALSE);
}


void
on_edit_profile_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
	if (!profile_dialog)
		profile_dialog = create_profile_dialog();
	gtk_widget_show_all(profile_dialog);
}


void
on_delete_profile_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
	if (get_selected_nr() > 0) {
		remove_profile_nr(get_selected_nr());
		config_sync();
	}
}


void
on_move_profile_up_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
	if (get_selected_nr() > 1) {
		swap_profile_nrs(get_selected_nr(), get_selected_nr() - 1);
		config_sync();
		set_selected_nr(get_selected_nr() - 1);
	}
}


void
on_move_profile_down_clicked           (GtkButton       *button,
                                        gpointer         user_data)
{
	if (get_selected_nr() > 0) {
		swap_profile_nrs(get_selected_nr(), get_selected_nr() + 1);
		config_sync();
		set_selected_nr(get_selected_nr() + 1);
	}
}


void
on_duplicate_profile_clicked           (GtkButton       *button,
                                        gpointer         user_data)
{
	if (get_selected_nr() >= 0) {
		gchar *oldname;
		gchar *newname;
		gint nr;

		if (!profile_dialog) {
			profile_dialog = create_profile_dialog();
			update_profile_dialog();
		}

		push_profile_nr(get_selected_nr());
		oldname = GET_STR(program.name);
		pop_profile();

		nr = push_new_profile();
		put_values(profile_dialog, profile_items);
		newname = g_strdup_printf("%s - copy", oldname);
		SET_STR(program.name, newname);
		SET_STR(program.new_win_key, "");
		SET_STR(program.new_tab_key, "");
		pop_profile();
		config_sync();

		g_free(oldname);
		g_free(newname);
	}
}


void
on_preferences_dialog_destroy          (GtkObject       *object,
                                        gpointer         user_data)
{
	preferences_dialog = NULL;
}

