/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
 * Copyright (c)2001,2002 Adrian Drzewiecki
 */
#include "dzt.h"

static gint pipe_fd = 0;

static void remote_read (gpointer data,
		   	 gint source,
			 GdkInputCondition condition);

static int lock		(int fd);
static int unlock	(int fd);

gint
remote_init()
{
	gchar pidfile_fmt[]  = "/%s/.gnome_private/dzt.pid.%s";
	gchar procinf_fmt[]  = "/proc/%s/cmdline";
	gchar pipefile_fmt[] = "/%s/.gnome_private/dzt.pipe.%s";

	gchar pidfile[512];
	gchar procinf[512];
	gchar pipefile[512];

	gchar tmpstr[512];

	gchar *home;
	gchar *display;

	FILE *fptr;
	FILE *fcmd;

	home = getenv("HOME");
	display = getenv("DISPLAY");

	snprintf(pipefile, sizeof(pipefile) - 1, pipefile_fmt, home, display);

	if (mkfifo(pipefile, S_IRWXU) < 0 && errno != EEXIST) {
		g_warning("failed to create pipe (%s): %s\n",
				pipefile, g_strerror(errno));
		return FALSE;
	}

	pipe_fd = open(pipefile, O_RDWR);

	snprintf(pidfile, sizeof(pidfile) - 1, pidfile_fmt, home, display);
	fptr = fopen(pidfile, "r+");
	if (!fptr) {
		fptr = fopen(pidfile, "w+");
		if (!fptr) {
			g_warning("failed to open %s: %s\n",
					pidfile, g_strerror(errno));
			return FALSE;
		}
	}

	if (!lock(fileno(fptr))) {
		g_warning("failed to lock %s: %s\n", pidfile,
				g_strerror(errno));
		fclose(fptr);
		return FALSE;
	}

	if (fscanf(fptr, "%s", tmpstr) > 0) {
		snprintf(procinf, sizeof(procinf) - 1, procinf_fmt, tmpstr);

		if (realpath(procinf, tmpstr)) {
			gchar *ptr;

			fcmd = fopen(procinf, "r");
			fscanf(fcmd, "%s", tmpstr);
			fclose(fcmd);

			ptr = strrchr(tmpstr, '/');
			if (ptr)
				ptr ++;
			else
				ptr = tmpstr;

			if (!strcmp(ptr, "dzt")) {
				unlock(fileno(fptr));
				fclose(fptr);
				return TRUE;
			}
		}
	}


	rewind(fptr);
	if (fprintf(fptr, "%d\n", getpid()) < 0) {
		g_warning("failed to update pid file: %s\n",
				g_strerror(errno));
	} else {
		gdk_input_add(pipe_fd, GDK_INPUT_READ,
				(GdkInputFunction)remote_read, NULL);
	}

	unlock(fileno(fptr));

	fclose(fptr);

	return FALSE;
}

gint
remote_start(gchar *prog)
{
	gchar prefix[] = "dzt ";
	gchar default_prog[] = "";
	gchar command[512];

	gchar *data;

	gint written;
	gint remaining;

	if (!prog)
		prog = default_prog;

	bzero(command, sizeof(command));
	snprintf(command, sizeof(command) - 1, "%s%s", prefix, prog);

	data = command;
	remaining = strlen(data);
	written = 0;

	if (!lock(pipe_fd)) {
		g_warning("failed to lock pipe: %s\n", g_strerror(errno));
		return FALSE;
	}

	while (remaining) {
		written = write(pipe_fd, data, remaining);
		if (written < 0) {
			g_warning("problem while writing to pipe: %s\n",
					g_strerror(errno));
			unlock(pipe_fd);
			return FALSE;
		}
		if (!written)
			break;

		data += written;
		remaining -= written;
	}

	unlock(pipe_fd);

	return TRUE;
}

static void
remote_read(gpointer data, gint source, GdkInputCondition condition)
{
	gchar buffer[512];
	gint  count;

	if (condition != GDK_INPUT_READ)
		return;

	bzero(buffer, sizeof(buffer));

	if (!lock(pipe_fd)) {
		g_warning("failed to lock pipe: %s\n", g_strerror(errno));
		return;
	}
	count = read(pipe_fd, buffer, sizeof(buffer));
	unlock(pipe_fd);

	if (count < 0)
		return;

	if (strncmp(buffer, "dzt", 3))
		return;

	//window_create_with_program(&buffer[4]);
	window_create(DEFAULT_PROFILE_ID);
}

static int
lock(int fd)
{
	struct flock	flk;
	int		ret;

	bzero(&flk, sizeof(flk));

	ret = fcntl(fd, F_GETLK, &flk);
	if (ret < 0)
		return 0;

	flk.l_type = F_WRLCK;

	for (;;) {
		ret = fcntl(fd, F_SETLKW, &flk);
		if (!ret)
			return 1;
		if (errno != EINTR)
			return 0;
	}
}

static int
unlock(int fd)
{
	struct flock	flk;
	int		ret;

	bzero(&flk, sizeof(flk));

	ret = fcntl(fd, F_GETLK, &flk);
	if (ret < 0)
		return 0;

	flk.l_type = F_UNLCK;
	
	ret = fcntl(fd, F_SETLK, &flk);
	if (ret < 0) {
		g_warning("Unlock of fd %d failed.\n", fd);
		return 0;
	}
	return 1;
}
