/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
 * Copyright (c)2001,2002 Adrian Drzewiecki
 */
#include "dzt.h"

typedef struct _WindowInfo WindowInfo;

struct _WindowInfo
{
	GtkWidget *notebook;
	GtkWidget *toolbar;
};

#define window_info(window)		GET_OBJECT_INFO(window, WindowInfo)
#define window_notebook(window)		(window_info(window)->notebook)
#define window_toolbar(window)		(window_info(window)->toolbar)

static inline GtkWidget *
toolbar_button_window(GtkWidget *button)
{
	while(!GTK_IS_WINDOW(button))
		button = button->parent;
	return button;
}

static GList *windows = NULL;

static GtkWidget *window_create_helper(void);

static void toolbar_add (GtkWidget *app,
		         gchar *icon,
			 gchar *text,
			 GtkSignalFunc action);

static gint on_window_destroy	   (GtkWidget *window);
static gint on_window_key_press	   (GtkWidget *window, GdkEventKey *ev);
static void on_window_configure	   (GtkWidget *window);
static void on_window_button_press (GtkWidget *window);

static void on_toolbar_new_tab_click    (GtkWidget *button);
static void on_toolbar_new_window_click (void);
static void on_toolbar_detach_tab_click (GtkWidget *button);
static void on_toolbar_rename_tab_click (GtkWidget *button);
static void on_toolbar_close_tab_click  (GtkWidget *button);
static void on_toolbar_prev_tab_click   (GtkWidget *button);
static void on_toolbar_next_tab_click   (GtkWidget *button);
static void on_toolbar_close_all_click	(void);

void
window_create(gint profileid)
{
	GtkWidget *app = window_create_helper();
	if (profileid != AUTOSTART) {
		page_create(window_notebook(app), profileid);
	} else {
		gint nr;
		for (nr = 0; push_profile_nr(nr) ;nr ++) {
			gboolean autostart = GET_BOOL(program.autostart);
			profileid = GET_INT(id);
			if (autostart || profileid == DEFAULT_PROFILE_ID)
				page_create(window_notebook(app), profileid);
			pop_profile();
		}
		g_print("\n");
	}
	window_configure(app);
}

void
window_create_with_page(GtkWidget *page)
{
	if (!page_is_lonely(page)) {
		GtkWidget *app = window_create_helper();
		page_moveto(page, window_notebook(app));
		window_configure(app);
	}
}

void
window_update_title(GtkWidget *window, gchar *title)
{
	if (window && title) {
		gchar newtitle[512];
		extern gchar dzt_program [];

		g_snprintf(newtitle, sizeof(newtitle) - 1, "%s - %s",
				dzt_program, title);
		gtk_window_set_title(GTK_WINDOW(window), newtitle);
	}
}

void
window_configure(GtkWidget *window)
{
	GtkToolbar *tb = GTK_TOOLBAR(window_toolbar(window));

	notebook_configure(window_notebook(window));

	if (GET_INT(main.toolbar_style) <= GTK_TOOLBAR_BOTH) {
		gtk_toolbar_set_style(tb, GET_INT(main.toolbar_style));
		gtk_widget_show(GTK_WIDGET(tb));
	} else
		gtk_widget_hide(GTK_WIDGET(tb));

	gtk_widget_show(window);
}

void
window_configure_all()
{
	g_list_foreach(windows, (GFunc)window_configure, NULL);
}

void
window_close(GtkWidget *window)
{
	gtk_widget_destroy(window);
}

void
window_close_all()
{
	if (GET_BOOL(main.confirm_close_all)) {
		GtkWidget *dialog = gnome_message_box_new(
			_("Are you sure you want to close all dzt windows?"),
			GNOME_MESSAGE_BOX_QUESTION,
			GNOME_STOCK_BUTTON_YES,
			GNOME_STOCK_BUTTON_NO,
			NULL);

		if (gnome_dialog_run_and_close(GNOME_DIALOG(dialog)) == 1)
			return;
	} 
	gtk_exit(0);
}

static GtkWidget *
window_create_helper()
{
	extern gchar dzt_program [];

	GtkWidget  *app  = gnome_app_new(dzt_program, dzt_program);
	GtkWidget  *box  = gtk_vbox_new(FALSE, 0);

	windows = g_list_append(windows, (gpointer)app);

	gtk_signal_connect(GTK_OBJECT(app), "destroy",
			GTK_SIGNAL_FUNC(on_window_destroy), NULL);
	gtk_signal_connect(GTK_OBJECT(app), "key-press-event",
			GTK_SIGNAL_FUNC(on_window_key_press), NULL);
	gtk_signal_connect(GTK_OBJECT(app), "configure-event",
			GTK_SIGNAL_FUNC(on_window_configure), NULL);
	gtk_signal_connect(GTK_OBJECT(app), "button-press-event",
			GTK_SIGNAL_FUNC(on_window_button_press), NULL);

	window_toolbar(app) = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL,
			GTK_TOOLBAR_BOTH);
	gtk_toolbar_set_button_relief(GTK_TOOLBAR(window_toolbar(app)),
			GTK_RELIEF_NONE);
	window_notebook(app) = notebook_create();

	gnome_app_set_contents(GNOME_APP(app), box);

	gtk_box_pack_start(GTK_BOX(box), window_toolbar(app), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(box), window_notebook(app), TRUE, TRUE, 0);

	toolbar_add(app,
		GNOME_STOCK_PIXMAP_EXEC,      
		"new tab",   
		GTK_SIGNAL_FUNC(on_toolbar_new_tab_click));
	toolbar_add(app,
		GNOME_STOCK_PIXMAP_NEW,       
		"new window",
		GTK_SIGNAL_FUNC(on_toolbar_new_window_click));
	toolbar_add(app,
		GNOME_STOCK_PIXMAP_UNDO,
		"detach tab",
		GTK_SIGNAL_FUNC(on_toolbar_detach_tab_click));
	toolbar_add(app,
		GNOME_STOCK_PIXMAP_PROPERTIES,
		"rename tab",
		GTK_SIGNAL_FUNC(on_toolbar_rename_tab_click));
	toolbar_add(app,
		GNOME_STOCK_PIXMAP_CLOSE,     
		"close tab", 
		GTK_SIGNAL_FUNC(on_toolbar_close_tab_click));
	toolbar_add(app,
		GNOME_STOCK_PIXMAP_BACK,      
		"prev tab",  
		GTK_SIGNAL_FUNC(on_toolbar_prev_tab_click));
	toolbar_add(app,
		GNOME_STOCK_PIXMAP_FORWARD,   
		"next tab",  
		GTK_SIGNAL_FUNC(on_toolbar_next_tab_click));
	toolbar_add(app,
		GNOME_STOCK_PIXMAP_EXIT,   
		"close all",  
		GTK_SIGNAL_FUNC(on_toolbar_close_all_click));

	return app;
}

static void
toolbar_add(GtkWidget *app, gchar *icon, gchar *text, GtkSignalFunc action)
{
	gtk_toolbar_append_item(GTK_TOOLBAR(window_toolbar(app)),
				_(text), _(text), "",
				gnome_stock_new_with_icon(icon), action, NULL);
}

static gint
on_window_destroy(GtkWidget *window)
{
	windows = g_list_remove(windows, (gpointer)window);
	if (!windows)
		gtk_exit(0);

	return TRUE;
}

static gint
on_window_key_press(GtkWidget *window, GdkEventKey *ev)
{
	if (menu_do_accel(window_notebook(window), ev->state, ev->keyval)) {
		ev->state  = 0;
		ev->keyval = 0;
		ev->length = 0;
		return TRUE;
	}

	return FALSE;
}

static void
on_window_configure(GtkWidget *window)
{
	gtk_widget_queue_draw(window_notebook(window));
}

static void 
on_window_button_press(GtkWidget *window)
{
	notebook_grab_focus(window_notebook(window));
}

static void
on_toolbar_new_tab_click(GtkWidget *button)
{
	GtkWidget *window   = toolbar_button_window(button);
	GtkWidget *notebook = window_notebook(window);

	page_create(notebook, DEFAULT_PROFILE_ID);
}

static void
on_toolbar_new_window_click()
{
	window_create(DEFAULT_PROFILE_ID);
}

static void
on_toolbar_detach_tab_click(GtkWidget *button)
{
	GtkWidget *window   = toolbar_button_window(button);
	GtkWidget *notebook = window_notebook(window);
	GtkWidget *page     = notebook_current_page(notebook);

	window_create_with_page(page);
}

static void
on_toolbar_rename_tab_click(GtkWidget *button)
{
	GtkWidget *window   = toolbar_button_window(button);
	GtkWidget *notebook = window_notebook(window);
	GtkWidget *page     = notebook_current_page(notebook);

	page_begin_rename(page, FALSE);
}

static void
on_toolbar_close_tab_click(GtkWidget *button)
{
	GtkWidget *window   = toolbar_button_window(button);
	GtkWidget *notebook = window_notebook(window);
	GtkWidget *page     = notebook_current_page(notebook);

	page_remove(page);
}

static void
on_toolbar_prev_tab_click(GtkWidget *button)
{
	notebook_go_prev_page(window_notebook(toolbar_button_window(button)));
}

static void
on_toolbar_next_tab_click(GtkWidget *button)
{
	notebook_go_next_page(window_notebook(toolbar_button_window(button)));
}

static void
on_toolbar_close_all_click()
{
	window_close_all();
}
